//
// $Id: cipher_one.cc 6084 2014-01-19 22:16:23Z dweber $
//
// simple substitution permutation cipher 
// cipher one from the block cipher companion,
// differential cryptanalysis chapter 6.1 on page 110
//
// the cipher has a
//    - 4-bit block size
//    - 8-bit key k (high part 4 bit k0, low part 4 bit k1)
//    - 4-bit S-BOX
//
//  c=S(m XOR k0) XOR k1
//
//  decomposing the steps via
//  u=m XOR k0
//  v=S(u)
//  c=v XOR k1
//
#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <stdlib.h>

#include "utils.h"  

using namespace std;

typedef enum { INVALID, ENCRYPT, DECRYPT, SBOX, BRUTE, DIFFERENTIAL} 
	cipher_mode_t; 

/*   --------------------------------------------------- */
BYTE sbox(BYTE b)
{
	const int SBOX_SIZE=16;
	const BYTE S[SBOX_SIZE]=
	{0x6,0x4,0xc,0x5,0x0,0x7,0x2,0xe,0x1,0xf,0x3,0xd,0x8,0xa,0x9,0xb};

	return S[b];
}

/*   --------------------------------------------------- */
BYTE sbox_inv(BYTE b)
{
	const int SBOX_SIZE=16;
	const BYTE S1[SBOX_SIZE]=
	{0x04,0x08,0x06,0x0a,0x01,0x03,0x00,0x05,
	 0x0c,0x0e,0x0d,0x0f,0x02,0x0b,0x07,0x09};

	return S1[b];
}


/*   --------------------------------------------------- */
void print_sbox()
{
	BYTE i;
	BYTE j;

	cout << "           ";
	for (i=0;i<16;i++)
		cout << hex << setw(2) << setfill('0') << (int) i << " ";
	cout << endl;

	cout << "sbox     : ";
	for (i=0;i<16;i++)
		cout << hex << setw(2) << setfill('0') << (int) sbox(i) << " ";
	cout << endl << endl;

	cout << "           ";
	for (i=0;i<16;i++)
		cout << hex << setw(2) << setfill('0') << (int) i << " ";
	cout << endl;

	cout << "sbox^{-1}: ";
	for (i=0;i<16;i++)
		for (j=0;j<=16;j++)
			if (sbox(j)==i)
			{
				cout << hex << setw(2) << setfill('0') << (int) j << " ";
				break;
			}
	cout << endl;
}

/*   --------------------------------------------------- */
/* only lower half of the BYTE gets encrypted */
//  8-bit key k (high part 4 bit k0, low part 4 bit k1)
//    
//  u=m XOR k0
//  v=S(u)
//  c=v XOR k1
BYTE encrypt(BYTE m, BYTE k)
{
	BYTE u,v,c;
	BYTE k0,k1;

	k0=(k>>4);
	k1=(k&0xf);
	u=m^k0;
	v=sbox(u);
	c=v^k1;

	return c;
}

// reverse encrypt:
//  v=c XOR k1
//  u=S^{-1}(v)
//  m=u XOR k0
BYTE decrypt(BYTE c, BYTE k)
{
	BYTE u,v;
	BYTE k0,k1;
	BYTE m;

	k0=(k>>4);
	k1=(k&0xf);
	v=c^k1;
	u=sbox_inv(v);
	m=u^k0;

	return m;
}

BYTE brute_force(BYTE m,BYTE c)
{
	BYTE key;  /* use key to loop over key space */ 
	BYTE rkey; /* returned key */

	rkey=key=0;
	do
	{
		if (encrypt(m,key)==c)
		{
			rkey=key; /* one of the right keys */
			cout << "key=" << std::hex << setw(2) << setfill('0') << (int) key << endl; 
		}
		key++;
	}
	while (key);
	return rkey;
}

BYTE differential(BYTE m0,BYTE m1,BYTE c0,BYTE c1)
{
	BYTE k1;  /* use k1 to loop over half the key length */ 
	BYTE rk1; /* return one of the candidates */
	BYTE md;

	md=m0^m1;
	cout << "k1 candidates, where sbox_inv(c0^k1)^sbox_inv(c1^k1)=m0^m1="<< hex << setw(2) << setfill('0') << (int) md << endl;

	rk1=k1=0;
	while (k1<=0xf)
	{
		if ((sbox_inv(c0^k1)^sbox_inv(c1^k1))==md)
		{
			cout << "k1 = " << hex << setw(2) << setfill('0') << (int) k1;
			cout << " => k0 = " << hex << setw(2) << setfill('0') << (int) (m0^sbox_inv(c0^k1)) << endl;
			rk1=k1;
		}
		k1++;
	}
	cout << "the k0 values have been computed via m0^sbox_inv(c0^k1)" << endl;
	return rk1;
}

int main(int argc,char **argv)
{
	BYTE data; // generic input data (plain or cipher)
	BYTE key;

	BYTE c; // cipher text
	BYTE m; // plain text

	BYTE m0,m1; // plain text pair
	BYTE c0,c1; // cipher text pair

	cipher_mode_t mode; // encrypt, decrypt, sbox info

	if (argc<=1 || (argc<=3 && argv[1][0]!='s'))
	{
		cout << "usage: " << argv[0] << " e|d <plain> <key>" << endl;
		cout << "usage: " << argv[0] << " s" << endl;
		cout << "         4-bit-values here" << endl;
		cout << "usage: " << argv[0] << " b <plain> <chiffre>" << endl;
		cout << "         brute force the key" << endl;
		cout << "usage: " << argv[0] << " a <m0> <m1> <c0> <c1>" << endl;
		cout << "         differential attack" << endl;
		cout << endl;
		cout << "for printing sbox table use" << endl;
		cout << "       " << argv[0] << " s" << endl;
		return 1;
	}

	mode=INVALID;
	if (argv[1][0]=='e') mode=ENCRYPT; // encrypt
	if (argv[1][0]=='d') mode=DECRYPT; // decrypt
	if (argv[1][0]=='s') mode=SBOX; // print sbox, sbox^{-1}
	if (argv[1][0]=='b') mode=BRUTE; // brute force the key
	if (argv[1][0]=='a') mode=DIFFERENTIAL; // differential attack
	if (mode==INVALID) 
	{
		cout << "unknown mode" << endl;
		return 1;
	}

	switch (mode)
	{
		case ENCRYPT:
		case DECRYPT:
			data=input_hex_byte(argv[2]);
			key=input_hex_byte(argv[3]);

			if (mode==ENCRYPT)
			{
				m=data;
				c=encrypt(m,key);
				cout << "encrypt m=" << std::hex << (int) m 
					<< " -> c=" << std::hex << (int) c 
					<< endl;
			}
			else
			{
				c=data;
				m=decrypt(c,key);
				cout << "decrypt c=" << std::hex << (int) c 
					<< " -> m=" << std::hex << (int) m 
					<< endl;
			}
			break;

		case BRUTE: 
				cout << "brute force" << endl;
				m=input_hex_byte(argv[2]);
				c=input_hex_byte(argv[3]);
				key=brute_force(m,c);
			      	return 0;

		case DIFFERENTIAL: 
				cout << "differential attack" << endl;
				m0=input_hex_byte(argv[2]);
				m1=input_hex_byte(argv[3]);
				c0=input_hex_byte(argv[4]);
				c1=input_hex_byte(argv[5]);
				key=differential(m0,m1,c0,c1);
			      	return 0;

		case INVALID: 
				cout << "invalid mode" << endl;
				return 1;

		case SBOX: print_sbox();
			   return 1;
	}
	return 0;
}

