//
// $Id: cipher_three.cc 6799 2014-03-30 08:11:23Z dweber $
//
// simple substitution permutation cipher 
// cipher THREE from the block cipher companion,
// differential cryptanalysis chapter 6.1 on page 110
//
// the cipher has a
//    - 4-bit block size
//    - 16-bit key k (four parts k0,k1,k2,k3 of size 4)
//    - 4-bit S-BOX
//
//
#include <iostream>
#include <fstream>
#include <iomanip>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "utils.h"  

using namespace std;

typedef enum { INVALID, ENCRYPT, DECRYPT, SBOX, BRUTE, 
		DIFFERENTIAL, SBOX_DIFFERENTIAL, CHOSEN_PLAINTEXT } 
	cipher_mode_t; 

/*   --------------------------------------------------- */
BYTE sbox(BYTE b)
{
	const int SBOX_SIZE=16;
	const BYTE S[SBOX_SIZE]=
	{0x6,0x4,0xc,0x5,0x0,0x7,0x2,0xe,0x1,0xf,0x3,0xd,0x8,0xa,0x9,0xb};

	return S[b];
}

/*   --------------------------------------------------- */
BYTE sbox_inv(BYTE b)
{
	const int SBOX_SIZE=16;
	const BYTE S1[SBOX_SIZE]=
	{0x04,0x08,0x06,0x0a,0x01,0x03,0x00,0x05,
		0x0c,0x0e,0x0d,0x0f,0x02,0x0b,0x07,0x09};

	return S1[b];
}

/*   --------------------------------------------------- */

void sbox2_difference_distribution()
{
#define SBOX_OUTPUT_VALUES 16
	BYTE b1,b2;
	BYTE d1,d2;
	BYTE d3,d3_inv;
	BYTE di;
	int output_diff_count[SBOX_OUTPUT_VALUES][SBOX_OUTPUT_VALUES];
	int output_diff_count2[SBOX_OUTPUT_VALUES][SBOX_OUTPUT_VALUES];
	int output_diff_count3[SBOX_OUTPUT_VALUES][SBOX_OUTPUT_VALUES];
	int output_diff_count3_inv[SBOX_OUTPUT_VALUES][SBOX_OUTPUT_VALUES];

	memset(output_diff_count,0,sizeof(int)*SBOX_OUTPUT_VALUES*SBOX_OUTPUT_VALUES);
	memset(output_diff_count2,0,sizeof(int)*SBOX_OUTPUT_VALUES*SBOX_OUTPUT_VALUES);
	memset(output_diff_count3,0,sizeof(int)*SBOX_OUTPUT_VALUES*SBOX_OUTPUT_VALUES);
	memset(output_diff_count3_inv,0,sizeof(int)*SBOX_OUTPUT_VALUES*SBOX_OUTPUT_VALUES);

	for (b1=0;(int) b1<SBOX_OUTPUT_VALUES;b1++)
		for (b2=0;(int) b2<SBOX_OUTPUT_VALUES;b2++)
		{
			d1=b1^b2;
			d2=sbox(b1)^sbox(b2);
			output_diff_count[d1][d2]++;
			d3=sbox(sbox(b1))^sbox(sbox(b2));
			output_diff_count3[d1][d3]++;
			d3_inv=sbox_inv(sbox_inv(b1))^sbox_inv(sbox_inv(b2));
			output_diff_count3_inv[d1][d3_inv]++;
		}

	cout << "            ";
	for (d1=0;(int) d1<SBOX_OUTPUT_VALUES;d1++)
		cout << "  " << hex << setw(2) << (int) d1;
	cout << endl;

	for (d1=0;(int) d1<SBOX_OUTPUT_VALUES;d1++) // the inputs
		for (d2=0;(int) d2<SBOX_OUTPUT_VALUES;d2++) // the outputs
			for (di=0;(int) di<SBOX_OUTPUT_VALUES;di++) // the intermediate values
				output_diff_count2[d1][d2]+=output_diff_count[d1][di]*output_diff_count[di][d2];

	for (d1=0;(int) d1<SBOX_OUTPUT_VALUES;d1++) // the inputs
	{
		cout << "hexval " << hex << setw(2) << setfill('0') << (int)d1 << " : ";
		for (d2=0;(int) d2<SBOX_OUTPUT_VALUES;d2++) // the outputs
			cout << " " << dec << setw(3) << setfill(' ') << (int) output_diff_count2[d1][d2];
		cout << endl;
	}
	cout << endl;
	cout << "table for applying s() twice" << endl;
	for (d1=0;(int) d1<SBOX_OUTPUT_VALUES;d1++) // the inputs
	{
		cout << "hexval " << hex << setw(2) << setfill('0') << (int)d1 << " : ";
		for (d2=0;(int) d2<SBOX_OUTPUT_VALUES;d2++) // the outputs
			cout << " " << dec << setw(3) << setfill(' ') << (int) output_diff_count3[d1][d2];
		cout << endl;
	}
	cout << endl;
	cout << "inverse table for applying s^(-1) twice" << endl;
	for (d1=0;(int) d1<SBOX_OUTPUT_VALUES;d1++) // the inputs
	{
		cout << "hexval " << hex << setw(2) << setfill('0') << (int)d1 << " : ";
		for (d2=0;(int) d2<SBOX_OUTPUT_VALUES;d2++) // the outputs
			cout << " " << dec << setw(3) << setfill(' ') << (int) output_diff_count3[d1][d2];
		cout << endl;
	}
}

/*   --------------------------------------------------- */
void print_sbox()
{
	BYTE i;
	BYTE j;

	cout << "           ";
	for (i=0;i<16;i++)
		cout << hex << setw(2) << setfill('0') << (int) i << " ";
	cout << endl;

	cout << "sbox     : ";
	for (i=0;i<16;i++)
		cout << hex << setw(2) << setfill('0') << (int) sbox(i) << " ";
	cout << endl << endl;

	cout << "           ";
	for (i=0;i<16;i++)
		cout << hex << setw(2) << setfill('0') << (int) i << " ";
	cout << endl;

	cout << "sbox^{-1}: ";
	for (i=0;i<16;i++)
		for (j=0;j<=16;j++)
			if (sbox(j)==i)
			{
				cout << hex << setw(2) << setfill('0') << (int) j << " ";
				break;
			}
	cout << endl;
}

/*   --------------------------------------------------- */
/* only lower half of BYTE m gets encrypted */
//  16-bit key k 
//  (highest part 4 bit k0, 
//  next lower part 4 bit k1 
//  next lower part 4 bit k2)
//  (lowest part 4 bit k3)
//    
//  u=m XOR k0
//  v=S(u)
//  w=v XOR k1
//  x=S(w)
//  y=x XOR k2
//  z=S(w)
//  c=z XOR k3
//
BYTE encrypt(BYTE m, WORD k)
{
	BYTE u,v,c; /* as in cipher ONE */
	BYTE w,x;   /* as in cipher TWO */
	BYTE y,z;   /* new in cipher THREE */
	BYTE k0,k1,k2,k3;

	k0=(k>>12)&0xf;
	k1=(k>>8)&0xf;
	k2=((k>>4)&0xf);
	k3=(k&0xf);

#if 0
	printf("using key %x %x %x %x\n",k0,k1,k2,k3);
#endif
	u=m^k0;
	v=sbox(u);
	w=v^k1;
	x=sbox(w);
	y=x^k2;
	z=sbox(y);
	c=z^k3;

	return c;
}

// reverse encrypt:
//  z=c^k3
//  y=S^{-1}(z)
//  x=y^k2
//  w=S^{-1}(x)
//  v=w XOR k1
//  u=S^{-1}(v)
//  m=u XOR k0
BYTE decrypt(BYTE c, WORD k)
{
	BYTE u,v; /* as in cipher_one */
	BYTE w,x; /* as in cipher TWO */
	BYTE y,z; /* new in cipher THREE */
	BYTE k0,k1,k2,k3;
	BYTE m;

	k0=(k>>12)&0xf;
	k1=(k>>8)&0xf;
	k2=((k>>4)&0xf);
	k3=(k&0xf);

	z=c^k3;
	y=sbox_inv(z);
	x=y^k2;
	w=sbox_inv(x);
	v=w^k1;
	u=sbox_inv(v);
	m=u^k0;

	return m;
}

WORD brute_force(BYTE m,BYTE c)
{
	WORD key;  /* use key to loop over key space */ 
	WORD rkey; /* returned key */

	rkey=key=0;
	do
	{
		if (encrypt(m,key)==c)
		{
			rkey=key; /* one of the right keys */
			cout << "key=" << std::hex << setw(2) << setfill('0') << (int) key << endl; 
		}
		key++;
	}
	while (key<=0xffff);
	return rkey;
}

/* perform differential attack by using PT/CT pairs from file */
BYTE differential(const char *filename, BYTE output_d)
{
	BYTE k3;  /* use k3 to loop over half the key length */ 
	BYTE rk3=0; /* return one of the candidates */
	BYTE m0,c0,m1,c1; // PT/CT pair 
	int m0i,c0i,m1i,c1i; // PT/CT pair read from file as int

	const BYTE K3_MAX=16;
	int count_k3[K3_MAX];
	int max_count_k3;

	ifstream fin(filename);

	if (!fin) 
	{
		cout << "file open error: " << filename << endl;
		return 0;
	}

	memset(count_k3,0,K3_MAX*sizeof(int));

	while (fin >> hex >> m0i >> hex >> c0i >> hex >> m1i >> hex >> c1i)
	{
		m0=(BYTE)(m0i&0xf);
		c0=(BYTE)(c0i&0xf);
		m1=(BYTE)(m1i&0xf);
		c1=(BYTE)(c1i&0xf);

		cout << "PT/CT pair" << endl;
		cout << hex << setw(2) << setfill('0') << (int) m0;
		cout << " ";
		cout << hex << setw(2) << setfill('0') << (int) c0;
		cout << " ";
		cout << hex << setw(2) << setfill('0') << (int) m1;
		cout << " ";
		cout << hex << setw(2) << setfill('0') << (int) c1;
		cout << endl;
		cout << "k3 candidates" << endl;

		rk3=k3=0;
		while (k3<=0xf)
		{
			if ((sbox_inv(c0^k3)^sbox_inv(c1^k3))==output_d)
			{
				cout << "k3 = " << hex << setw(2) << setfill('0') << (int) k3 << endl;
				rk3=k3;
				count_k3[k3]++;
			}
			k3++;
		}
	}

	max_count_k3=0;
	for (k3=0;k3<K3_MAX;k3++)
	{
		cout << "count[" << hex << setw(2) << setfill('0') 
			<< (int) k3 << "]=" << count_k3[k3] << endl;
		if (count_k3[k3] > max_count_k3)
		{
			max_count_k3=count_k3[k3];
			rk3=k3;
		}
	}
	return rk3;
}

void prepare_chosen_plaintext(WORD key,BYTE d, BYTE out_d,int count,const char *filename)
{
	BYTE m0,m1; // messages
	BYTE c0,c1; // ciphertexts

	ofstream fout(filename);

	if (fout)
		cout << "writing PT/CT pairs to " << filename << endl;

	m0=0;
	while (((int)m0<count) && (m0<=0xf))
	{
		c0=encrypt(m0,key);
		m1=m0^d;
		c1=encrypt(m1,key);

		fout << hex << setw(2) << setfill('0') << (int) m0 
		<< " " 
		<< hex << setw(2) << setfill('0') << (int) c0 
		<< " " 
		<< hex << setw(2) << setfill('0') << (int) m1 
		<< " " 
		<< hex << setw(2) << setfill('0') << (int) c1 << endl;

		cout << hex << setw(2) << setfill('0') << (int) m0 
		<< " " 
		<< hex << setw(2) << setfill('0') << (int) c0 
		<< " " 
		<< hex << setw(2) << setfill('0') << (int) m1 
		<< " " 
		<< hex << setw(2) << setfill('0') << (int) c1;

		if ((c0^c1) == out_d)
			cout << "*" << endl;
		else
			cout << endl;


		m0++;
	}
}

int main(int argc,char **argv)
{
	BYTE data; // generic input data (plain or cipher)
	WORD key;

	BYTE c; // cipher text
	BYTE m; // plain text

	BYTE input_d; // input difference of S-box
	BYTE output_d; // output difference of S-box

	int count; // counting PT/CT-pairs

	char *filename; // output file for PT/CT attack

	cipher_mode_t mode; // encrypt, decrypt, sbox info

	if ((argc<=1 || (argc<=3 && argv[1][0]!='s' && argv[1][0]!='x')))
	{
		cout << "usage: " << argv[0] << " e|d <plain> <key>" << endl;
		cout << "usage: " << argv[0] << " s" << endl;
		cout << "usage: " << argv[0] << " b <plain> <chiffre>" << endl;
		cout << "         brute force the key" << endl;
		cout << "usage: " << argv[0] << " a <input filename> <input_d>" << endl;
		cout << "         differential attack on the cipher" << endl;
		cout << "usage: " << argv[0] << " x " << endl;
		cout << "         differential attack on the S-box" << endl;
		cout << "usage: " << argv[0] << " c <key> <d> <count> <output filename>" << endl;
		cout << "         prepare chosen plaintext differential attack" << endl;
		cout << endl;
		cout << "for printing sbox table use" << endl;
		cout << "       " << argv[0] << " s" << endl;
		cout << "         4-bit-values here" << endl;
		return 1;
	}

	mode=INVALID;
	if (argv[1][0]=='e') mode=ENCRYPT; // encrypt
	if (argv[1][0]=='d') mode=DECRYPT; // decrypt
	if (argv[1][0]=='s') mode=SBOX; // print sbox, sbox^{-1}
	if (argv[1][0]=='b') mode=BRUTE; // brute force the key
	if (argv[1][0]=='a') mode=DIFFERENTIAL; // differential attack
	if (argv[1][0]=='x') mode=SBOX_DIFFERENTIAL; // differential attack on S-box
	if (argv[1][0]=='c') mode=CHOSEN_PLAINTEXT; 
	// prepare chosen plaintext attack with given difference for plaintexts
	if (mode==INVALID) 
	{
		cout << "unknown mode" << endl;
		return 1;
	}

	switch (mode)
	{
		case ENCRYPT:
		case DECRYPT:
			data=input_hex_byte(argv[2]);
			key=input_hex_word(argv[3]);

			if (mode==ENCRYPT)
			{
				m=data;
				c=encrypt(m,key);
				cout << "encrypt m=" << std::hex << (int) m 
					<< " -> c=" << std::hex << (int) c 
					<< " (key=" << std::hex << (int) key
					<< ")"
					<< endl;
			}
			else
			{
				c=data;
				m=decrypt(c,key);
				cout << "decrypt c=" << std::hex << (int) c 
					<< " -> m=" << std::hex << (int) m 
					<< endl;
			}
			break;

		case BRUTE: 
				cout << "brute force" << endl;
				m=input_hex_byte(argv[2]);
				c=input_hex_byte(argv[3]);
				key=brute_force(m,c);
			      	return 0;

		case SBOX_DIFFERENTIAL: 
				cout << "difference distribution tables S-Box" << endl;
				sbox2_difference_distribution();
				return 0;
		case DIFFERENTIAL: 
				cout << "differential attack" << endl;
				filename=argv[2];
				output_d=input_hex_byte(argv[3]);
				key=differential(filename,output_d);
				cout << "differential attack k3=" << hex 
					<< (int) key << endl;
			      	return 0;

		case CHOSEN_PLAINTEXT: 
				cout << "prepare chosen plaintext attack with differential" << endl;
				key=input_hex_word(argv[2]);
				input_d=input_hex_word(argv[3]);
				output_d=input_hex_word(argv[4]);
				count=atoi(argv[5]);
				filename=argv[6];
				prepare_chosen_plaintext(key,input_d, output_d,count,filename);
			      	return 0;

		case INVALID: 
				cout << "invalid mode" << endl;
				return 1;

		case SBOX: print_sbox();
			   return 1;
	}
	return 0;
}

