//
// $Id: cipher_two.cc 7103 2014-04-22 16:41:29Z dweber $
//
// simple substitution permutation cipher 
// cipher TWO from the block cipher companion,
// differential cryptanalysis chapter 6.1 on page 110
//
// the cipher has a
//    - 4-bit block size
//    - 12-bit key k (high part 4 bit k0, low part 4 bit k1)
//    - 4-bit S-BOX
//
//
#include <iostream>
#include <fstream>
#include <iomanip>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "utils.h"  

using namespace std;

typedef enum { INVALID, ENCRYPT, DECRYPT, SBOX, BRUTE, 
		DIFFERENTIAL, SBOX_DIFFERENTIAL, CHOSEN_PLAINTEXT
		 } 
	cipher_mode_t; 

/*   --------------------------------------------------- */
#if 1
BYTE sbox(BYTE b)
{
	const int SBOX_SIZE=16;
	const BYTE S[SBOX_SIZE]=
	{0x6,0x4,0xc,0x5,0x0,0x7,0x2,0xe,0x1,0xf,0x3,0xd,0x8,0xa,0x9,0xb};

	return S[b];
}

/*   --------------------------------------------------- */
BYTE sbox_inv(BYTE b)
{
	const int SBOX_SIZE=16;
	const BYTE S1[SBOX_SIZE]=
	{0x04,0x08,0x06,0x0a,0x01,0x03,0x00,0x05,
		0x0c,0x0e,0x0d,0x0f,0x02,0x0b,0x07,0x09};

	return S1[b];
}
#else /* alternative S-Box */
BYTE sbox(BYTE b)
{
	const int SBOX_SIZE=16;
	const BYTE S[SBOX_SIZE]=
	{0xe,0x2,0x1,0x3,0xd,0x9,0x0,0x6,0xf,0x4,0x5,0xa,0x8,0xc,0x7,0xb};

	return S[b];
}

/*   --------------------------------------------------- */
BYTE sbox_inv(BYTE b)
{
	const int SBOX_SIZE=16;
	const BYTE S1[SBOX_SIZE]=
	{0x06,0x02,0x01,0x03,0x09,0x0a,0x07,0x0e,
		0x0c,0x05,0x0b,0x0f,0x0d,0x04,0x00,0x08};

	return S1[b];
}

#endif

/*   --------------------------------------------------- */

void sbox_difference_distribution()
{
#define SBOX_OUTPUT_VALUES 16
	BYTE b1,b2;
	BYTE d1,d2;
	int output_diff_count[SBOX_OUTPUT_VALUES][SBOX_OUTPUT_VALUES];

	memset(output_diff_count,0,sizeof(int)*SBOX_OUTPUT_VALUES*SBOX_OUTPUT_VALUES);

	for (b1=0;(int) b1<SBOX_OUTPUT_VALUES;b1++)
		for (b2=0;(int) b2<SBOX_OUTPUT_VALUES;b2++)
		{
			d1=b1^b2;
			d2=sbox(b1)^sbox(b2);
			output_diff_count[d1][d2]++;
		}

	cout << "     output ";
	for (d1=0;(int) d1<SBOX_OUTPUT_VALUES;d1++)
		cout << " " << hex << setw(2) << (int) d1;
	cout << endl;

	for (d1=0;(int) d1<SBOX_OUTPUT_VALUES;d1++)
	{
		cout << "hexval " << hex << setw(2) << setfill('0') << (int)d1 << " : ";
		for (d2=0;(int) d2<SBOX_OUTPUT_VALUES;d2++)
			cout << " " << dec << setw(2) << setfill(' ') << (int) output_diff_count[d1][d2];
		cout << endl;
	}

	cout << "a row (hexval) describes the input difference" << endl;
	cout << "a column describes the output difference" << endl;
}

/*   --------------------------------------------------- */
/* we check, whether a0x0+a1x1+a2x2+a3x3=b0y0+b1y1+b2y2+b3y3 mod 2 */
/* this is the same as
 * xor all bits of (a & x)   =   xor all bits of (b & y)
 */
static BYTE xorbits(BYTE b)
{
	BYTE xorsum=0;
	int i;

	for (i=0;i<4;i++)
	{
		xorsum^=(b&1);
		b>>=1;
	}
	return xorsum;
}



void print_binary(BYTE x)
{
	int i;

	for (i=3;i>=0;i--)
		if ((x & (0x1 << i))==0)
			cout << " 0";
		else
			cout << " 1";
}

void print_bit(BYTE x)
{
	if ((x&1)==0)
		cout << " 0";
	else
		cout << " 1";
}

void print_row(BYTE x, BYTE y, BYTE a, BYTE b)
{
	BYTE ax, by;

	print_binary(x);
	cout << "   ";
	print_binary(y);
	cout << "   ";

	ax=(a & x) & 0xf;
	by=(b & y) & 0xf;

	print_bit(xorbits(ax));
	cout << "   ";
	print_bit(xorbits(by));
	cout << "   ";

	if (xorbits(ax)==xorbits(by))
		cout << "   OK";

	cout << endl;
	
}


/*   --------------------------------------------------- */
void print_sbox()
{
	BYTE i;
	BYTE j;

	cout << "           ";
	for (i=0;i<16;i++)
		cout << hex << setw(2) << setfill('0') << (int) i << " ";
	cout << endl;

	cout << "sbox     : ";
	for (i=0;i<16;i++)
		cout << hex << setw(2) << setfill('0') << (int) sbox(i) << " ";
	cout << endl << endl;

	cout << "           ";
	for (i=0;i<16;i++)
		cout << hex << setw(2) << setfill('0') << (int) i << " ";
	cout << endl;

	cout << "sbox^{-1}: ";
	for (i=0;i<16;i++)
		for (j=0;j<=16;j++)
			if (sbox(j)==i)
			{
				cout << hex << setw(2) << setfill('0') << (int) j << " ";
				break;
			}
	cout << endl;
}

/*   --------------------------------------------------- */
/* only lower half of the BYTE gets encrypted */
//  8-bit key k 
//  (high part 4 bit k0, middle part 4 bit k1 low part 4 bit k2)
//    
//  u=m XOR k0
//  v=S(u)
//  w=v XOR k1
//  x=S(w)
//  c=x XOR k2
BYTE encrypt(BYTE m, WORD k)
{
	BYTE u,v,c; /* as in cipher_one */
	BYTE w,x;
	BYTE k0,k1,k2;

	k0=(k>>8)&0xf;
	k1=((k>>4)&0xf);
	k2=(k&0xf);

	u=m^k0;
	v=sbox(u);
	w=v^k1;
	x=sbox(w);
	c=x^k2;

	return c;
}

// reverse encrypt:
//  x=c^k2
//  w=S^{-1}(x)
//  v=w XOR k1
//  u=S^{-1}(v)
//  m=u XOR k0
BYTE decrypt(BYTE c, WORD k)
{
	BYTE u,v; /* as in cipher_one */
	BYTE w,x;
	BYTE k0,k1,k2;
	BYTE m;

	k0=(k>>8);
	k1=((k>>4)&0xf);
	k2=(k&0xf);

	x=c^k2;
	w=sbox_inv(x);
	v=w^k1;
	u=sbox_inv(v);
	m=u^k0;

	return m;
}

WORD brute_force(BYTE m,BYTE c)
{
	WORD key;  /* use key to loop over key space */ 
	WORD rkey; /* returned key */

	rkey=key=0;
	do
	{
		if (encrypt(m,key)==c)
		{
			rkey=key; /* one of the right keys */
			cout << "key=" << std::hex << setw(2) << setfill('0') << (int) key << endl; 
		}
		key++;
	}
	while (key<=0xfff);
	return rkey;
}

BYTE differential(const char *filename, BYTE output_d)
{
	BYTE k2;  /* use k2 to loop over half the key length */ 
	BYTE rk2=0; /* return one of the candidates */
	BYTE m0,c0,m1,c1; // PT/CT pair 
	int m0i,c0i,m1i,c1i; // PT/CT pair read from file as int

	const BYTE K2_MAX=16;
	int count_k2[K2_MAX];
	int max_count_k2;

	ifstream fin(filename);

	if (!fin) 
	{
		cout << "file open error: " << filename << endl;
		return 0;
	}

	memset(count_k2,0,K2_MAX*sizeof(int));

	while (fin >> hex >> m0i >> hex >> c0i >> hex >> m1i >> hex >> c1i)
	{
		m0=(BYTE)(m0i&0xf);
		c0=(BYTE)(c0i&0xf);
		m1=(BYTE)(m1i&0xf);
		c1=(BYTE)(c1i&0xf);

		cout << "PT/CT pair" << endl;
		cout << hex << setw(2) << setfill('0') << (int) m0;
		cout << " ";
		cout << hex << setw(2) << setfill('0') << (int) c0;
		cout << " ";
		cout << hex << setw(2) << setfill('0') << (int) m1;
		cout << " ";
		cout << hex << setw(2) << setfill('0') << (int) c1;
		cout << endl;
		cout << "k2 candidates" << endl;

		rk2=k2=0;
		while (k2<=0xf)
		{
			if ((sbox_inv(c0^k2)^sbox_inv(c1^k2))==output_d)
			{
				cout << "k2 = " << hex << setw(2) << setfill('0') << (int) k2 << endl;
				rk2=k2;
				count_k2[k2]++;
			}
			k2++;
		}
	}

	max_count_k2=0;
	for (k2=0;k2<K2_MAX;k2++)
	{
		cout << "count[" << hex << setw(2) << setfill('0') 
			<< (int) k2 << "]=" << dec << count_k2[k2] << endl;
		if (count_k2[k2] > max_count_k2)
		{
			max_count_k2=count_k2[k2];
			rk2=k2;
		}
	}

	cout << "candidates:" << endl;
	for (k2=0;k2<K2_MAX;k2++)
		if (count_k2[k2] == max_count_k2)
			cout << hex << setw(2) << setfill('0') << (int) k2 << endl;
	cout << endl;

	return rk2;
}

void prepare_chosen_plaintext(WORD key,BYTE d,int count,const char *filename)
{
	BYTE m0,m1; // messages
	BYTE c0,c1; // ciphertexts

	ofstream fout(filename);

	if (fout)
		cout << "writing PT/CT pairs to " << filename << endl;

	m0=0;
	while (((int)m0<count) && (m0<=0xf))
	{
		c0=encrypt(m0,key);
		m1=m0^d;
		c1=encrypt(m1,key);

		fout << hex << setw(2) << setfill('0') << (int) m0 
		<< " " 
		<< hex << setw(2) << setfill('0') << (int) c0 
		<< " " 
		<< hex << setw(2) << setfill('0') << (int) m1 
		<< " " 
		<< hex << setw(2) << setfill('0') << (int) c1 << endl;

		cout << hex << setw(2) << setfill('0') << (int) m0 
		<< " " 
		<< hex << setw(2) << setfill('0') << (int) c0 
		<< " " 
		<< hex << setw(2) << setfill('0') << (int) m1 
		<< " " 
		<< hex << setw(2) << setfill('0') << (int) c1 << endl;

		m0++;
	}
}

int main(int argc,char **argv)
{
	BYTE data; // generic input data (plain or cipher)
	WORD key;

	BYTE c; // cipher text
	BYTE m; // plain text

	BYTE input_d; // difference between plaintexts
	BYTE output_d; // difference between plaintexts


	int count; // counting PT/CT-pairs

	char *filename; // output file for PT/CT attack

	cipher_mode_t mode; // encrypt, decrypt, sbox info

	if (argc<=1 || (argc<=3 && argv[1][0]!='s' && argv[1][0]!='x' && argv[1][0]!='l'))
	{
		cout << "usage: " << argv[0] << " e|d <plain> <key>" << endl;
		cout << "usage: " << argv[0] << " s" << endl;
		cout << "usage: " << argv[0] << " b <plain> <chiffre>" << endl;
		cout << "         brute force the key" << endl;
		cout << "usage: " << argv[0] << " a <input filename> <output_d>" << endl;
		cout << "         differential attack on the cipher" << endl;
		cout << "usage: " << argv[0] << " x " << endl;
		cout << "         print difference distribution table" << endl;
		cout << "usage: " << argv[0] << " c <key> <d> <count> <output filename>" << endl;
		cout << "         prepare chosen plaintext differential attack" << endl;
		cout << endl;
		cout << "for printing sbox table use" << endl;
		cout << "       " << argv[0] << " s" << endl;
		cout << "         4-bit-values here" << endl;
		return 1;
	}

	mode=INVALID;
	if (argv[1][0]=='e') mode=ENCRYPT; // encrypt
	if (argv[1][0]=='d') mode=DECRYPT; // decrypt
	if (argv[1][0]=='s') mode=SBOX; // print sbox, sbox^{-1}
	if (argv[1][0]=='b') mode=BRUTE; // brute force the key
	if (argv[1][0]=='a') mode=DIFFERENTIAL; // differential cryptanalysis
	if (argv[1][0]=='x') mode=SBOX_DIFFERENTIAL; // differential attack on S-box
	if (argv[1][0]=='c') mode=CHOSEN_PLAINTEXT; 
	// prepare chosen plaintext attack with given difference for plaintexts
	if (mode==INVALID) 
	{
		cout << "unknown mode" << endl;
		return 1;
	}

	switch (mode)
	{
		case ENCRYPT:
		case DECRYPT:
			data=input_hex_byte(argv[2]);
			key=input_hex_word(argv[3]);

			if (mode==ENCRYPT)
			{
				m=data;
				c=encrypt(m,key);
				cout << "encrypt m=" << std::hex << (int) m 
					<< " -> c=" << std::hex << (int) c 
					<< endl;
			}
			else
			{
				c=data;
				m=decrypt(c,key);
				cout << "decrypt c=" << std::hex << (int) c 
					<< " -> m=" << std::hex << (int) m 
					<< endl;
			}
			break;

		case BRUTE: 
				cout << "brute force" << endl;
				m=input_hex_byte(argv[2]);
				c=input_hex_byte(argv[3]);
				key=brute_force(m,c);
			      	return 0;

		case SBOX_DIFFERENTIAL: 
				cout << "differential distribution table for S-Box" << endl;
				sbox_difference_distribution();
				return 0;
		case DIFFERENTIAL: 
				cout << "differential attack" << endl;
				filename=argv[2];
				output_d=input_hex_byte(argv[3]);
				key=differential(filename,output_d);
				cout << "differential attack k2=" << hex 
					<< (int) key << endl;
			      	return 0;

		case CHOSEN_PLAINTEXT: 
				cout << "prepare chosen plaintext attack with differential" << endl;
				key=input_hex_word(argv[2]);
				input_d=input_hex_word(argv[3]);
				count=atoi(argv[4]);
				filename=argv[5];
				prepare_chosen_plaintext(key,input_d,count,filename);
			      	return 0;

		case INVALID: 
				cout << "invalid mode" << endl;
				return 1;

		case SBOX: print_sbox();
			   return 1;
	}
	return 0;
}

